import sys,os,stat,time

import rox
from rox import saving,filer,mime,Menu,InfoWin
import gtk

import urlparse,urllib
from threading import Thread

import options, downloader
from downloader import Downloader
from support import *

class Progress(object):
	def __init__(self):
		self.styles = {
			'percent':self._percent,
			'speed':self._speed,
			'eta':self._eta,
			'size':self._size,
			'url':self._file,
		}
		self.string = APP_NAME
		
	def format(self,style,*args):
		self.string = self.styles[style](*args)
		
	def _percent(self,*args):
		total,count = map(float,self.byte_offset(args[1]))
		if total in [0,downloader.SIZE_UNKNOWN]:
			complete = 0
		else:
			complete = (count/total)*100
		return '%.2f%% Complete' % complete
		
	def _speed(self,*args):
		return '%d Bps' % self.calc_bps(*args[1:])
			
	def _eta(self,*args):
		total,count = self.byte_offset(args[1])
		bps = self.calc_bps(*args[1:])
			
		if 0 in [total,bps]:
			eta = 0
		else:
			remain = total - count
			eta = remain / bps
		return 'ETA: %d seconds' % eta
		
	def _size(self,*args):
		total,count = self.byte_offset(args[1])
		return '%d/%d' % (count,total)
		
	def _file(self,*args):
		url,headers = args[0]
		return url
		
	def byte_offset(self,byte_data):
		start,total,count = byte_data
		return (start+total,start+count)
		
	def calc_bps(self,byte_data,time_data):
		total,count = self.byte_offset(byte_data)
		start,now = time_data
		time_delta = now - start
		try:
			return count / time_delta.seconds
		except ZeroDivisionError:
			return 0
			
class UIThread(Thread):
	def __init__(self,sibling):
		Thread.__init__(self)
		self.sibling = sibling
	def run(self):
		while self.sibling.isAlive():
			time.sleep(0.1)
			gtk.threads_enter()
			run_main()
			gtk.threads_leave()

class TitleThread(Thread):
	def __init__(self,sibling,title_fcn):
		Thread.__init__(self)
		self.sibling = sibling
		self.title_fcn = title_fcn
	def run(self):
		sleep = options.title_freq.int_value
		while self.sibling.isAlive():
			time.sleep(sleep)
			gtk.threads_enter()
			self.title_fcn()
			gtk.threads_leave()
		
class DownloadThread(Thread):
	def __init__(self,download,stream,callback):
		Thread.__init__(self)
		self.download = download
		self.stream = stream
		self.callback = callback
	def run(self):
		self.download.write(self.stream,self.callback)

class ValidateThread(Thread):
	def __init__(self,download):
		Thread.__init__(self)
		self.download = download
	def run(self):
		self.download.validate()
		
class SaveBox(saving.SaveBox,saving.Saveable):
	def __init__(self):
		self.count = 0
		self.download = Downloader()
		self.file = ''
		self.old_file = ''
		
		self.options_box = WIDGETS.get_widget('thebox')
		self.entry_url = WIDGETS.get_widget('entry_url')
		self.button_validate = WIDGETS.get_widget('button_validate')
		self.button_options = WIDGETS.get_widget('button_options')
		
		saving.SaveBox.__init__(self,self,'','text/plain')
		self.save_area.confirm_new_path = self.confirm_new_path
		self.save_area.save_cancelled = self.save_cancelled
		
		self.set_title('ROXget')
		self.build_menu()
		
	def build_main_area(self):
		self.vbox.add(self.save_area)
		
		self.options_box.unparent()
		self.vbox.pack_start(self.options_box,False,True,0)
		self.options_box.show()
		
		self.entry_url.connect('changed',self.url_changed)
		self.button_validate.connect('clicked',self.validate_clicked)
		self.button_options.connect('clicked',self.options_clicked)
	
	def run(self):
		if options.validate.int_value:
			self.validate_clicked()
		
	def save_to_stream(self,stream):
		assert self.url, ERR_NOURL
		
		if self.download.valid != True:
			self.validate_clicked()
			
		if self.download.resume:
			self.old_file = self.file+'.part'
			os.rename(self.file,self.old_file)
		
		self.progress = Progress()
		
		self.set_title('Connecting')
		
		downloadthread = DownloadThread(self.download,stream,self.update_status)
		#downloadthread = ValidateThread(self.download)
		uithread = UIThread(downloadthread)
		titlethread = TitleThread(downloadthread,self.update_title)
		
		downloadthread.start()
		uithread.start()
		titlethread.start()
		titlethread.join()
		uithread.join()
		downloadthread.join()
		
		self.update_title()
		self.result = self.download.status
	
	def save_cancelled(self):
		if rox.confirm('Really abort download?',gtk.STOCK_YES):
			self.set_title('Closing session')
			self.download.abort()
	
	def set_save_in_progress(self,busy):
		self.button_options.set_sensitive(not busy)
		self.options_box.set_sensitive(not busy)
		saving.SaveBox.set_save_in_progress(self,busy)
	
	def set_uri(self,uri):
		self.file = uri
		if self.old_file:
			open(self.old_file,'ab').write(open(self.file).read())
			os.rename(self.old_file,self.file)
		if self.result == downloader.SUCCESSFUL:
			complete = CompleteDialog(self.file).run()
			if complete == gtk.RESPONSE_YES:
				#execute
				filer.spawn_rox((self.file,))
			elif complete == gtk.RESPONSE_NO:
				# show
				filer.show_file(self.file)
		else:
			strings = {
				downloader.INCOMPLETE:'Premature EOF.',
				downloader.ABORTED:'Aborted by user.',
				downloader.UNKNOWN_ERROR:'Unhandled error encountered.',
			}
			if rox.confirm(strings[self.result]+'\n\nDelete file %s'%uri,gtk.STOCK_DELETE):
				os.remove(uri)
			
	
	def set_url(self,url):
		self.entry_url.set_text(url)

	def confirm_new_path(self, path):
		if not os.path.exists(path):
			return True
		
		resume = ResumeDialog(path).run()
		if resume == gtk.RESPONSE_YES:
			if self.download.valid != True:
				self.validate_clicked()
			self.download.resume_from(os.stat(path)[stat.ST_SIZE])
			return True
		if resume == gtk.RESPONSE_NO:
			return True
	
	def url_changed(self,*args):
		self.button_validate.set_sensitive(True)
		self.url = self.entry_url.get_text()
		parts = urlparse.urlparse(self.url)
		
		file = urllib.unquote(parts[2].split('/')[-1])
		if not file:
			file = 'default.html'
		
		path = os.path.split(self.save_area.entry.get_text())[0]
		if not path:
			path = os.path.abspath(os.path.curdir)
		
		abspath = os.path.join(path,file)
		self.save_area.entry.set_text(abspath)
		
		start = abspath.rfind('/')+1
		gtk.Editable.select_region(self.save_area.entry,start,-1)
		
		self.set_mime(file)
		self.download.url = self.url
		self.file = abspath
	
	def validate_clicked(self,*args):
		self.set_title('Validating')
		self.set_sensitive(False)
		
		validatethread = ValidateThread(self.download)
		uithread = UIThread(validatethread)
		validatethread.start()
		uithread.start()
		uithread.join()
		validatethread.join()
		
		result = self.download.valid
		if result == True:
			self.set_url(self.download.url)
			self.button_validate.set_sensitive(False)
		else:
			rox.alert(result)
		self.set_title('ROXget')
		self.set_sensitive(True)
	
	def options_clicked(self,*args):
		rox.edit_options()
		
	def info_clicked(self,*args):
		InfoWin.infowin(APP_NAME)
		
	def quit_clicked(self,*args):
		if self.download.status == downloader.RUNNING:
			self.save_cancelled()
		self.destroy()
		
	def set_mime(self,file):
		try:
			self.set_type(self.download.fp.headers['content-type'])
		except:
			m,s = ('text','plain')
			t = mime.get_type_by_name(file)
			if t:
				m = t.media
				s = t.subtype
			self.set_type('/'.join((m,s)))
	
	def update_status(self,*args):
		self.progress.format(options.progress_style.value,*args)
		
	def update_title(self):
		self.set_title(self.progress.string)
		
	def build_menu(self):
		self.add_events(gtk.gdk.BUTTON_PRESS_MASK)
		self.connect('button-press-event',self.button_press)
		
		menu_main = [
			#Menu.Action('Status','not_done',),
			#Menu.Action('Log','not_done'),
			#Menu.Separator(),
			Menu.Action('Options','options_clicked',stock=gtk.STOCK_PREFERENCES),
			Menu.Action('Info','info_clicked',stock=gtk.STOCK_DIALOG_INFO),
			Menu.Separator(),
			Menu.Action('Quit','quit_clicked',stock=gtk.STOCK_QUIT),
			]
			
		rox.Menu.set_save_name(APP_NAME)
		self.menu = Menu.Menu('main',menu_main)
		
	def button_press(self,window,event):
		if event.button == 3:
			self.menu.popup(self,event)

class CompleteDialog:
	def __init__(self,path):
		self.dialog = WIDGETS.get_widget('complete_dialog')
		self.filename = WIDGETS.get_widget('c_filename')
		self.execute = WIDGETS.get_widget('button_execute')
		self.show = WIDGETS.get_widget('button_show')
		self.close = WIDGETS.get_widget('button_close')
		
		self.dialog.connect('delete_event',self.destroy)
		self.execute.connect('clicked',self.destroy)
		self.show.connect('clicked',self.destroy)
		self.close.connect('clicked',self.destroy)
		
		self.filename.set_text(path)
		
		self.finish_timeout = options.finish_timeout.int_value
		if self.finish_timeout:
			self.timeout = gtk.timeout_add(self.finish_timeout*1000,self.destroy)
		
	def destroy(self,*args):
		self.dialog.hide()
		if self.finish_timeout:
			gtk.timeout_remove(self.timeout)
		return gtk.TRUE
	
	def run(self):
		return self.dialog.run()
		
class ResumeDialog:
	def __init__(self,path):
		self.dialog = WIDGETS.get_widget('resume_dialog')
		self.filename = WIDGETS.get_widget('r_filename')
		self.resume = WIDGETS.get_widget('button_resume')
		self.overwrite = WIDGETS.get_widget('button_overwrite')
		self.cancel = WIDGETS.get_widget('button_cancel')
		
		self.dialog.connect('delete_event',self.destroy)
		self.resume.connect('clicked',self.destroy)
		self.overwrite.connect('clicked',self.destroy)
		self.cancel.connect('clicked',self.destroy)
		
		self.filename.set_text(path)
	
	def destroy(self,*args):
		self.dialog.hide()
		return gtk.TRUE
	
	def run(self):
		return self.dialog.run()
